#include "config_private.h"
#include "vm.h"
#include "program.h"
#include "stdlib.h"
#include "funcsig.h"
#include "byteorder.h"
#include "function.h"
#include "loader.h"
#include "processor.h"
#include "routines.h"
#include "type.h"
#include "operator.h"
#include "vector.h"
#include "ctx.h"
#include <libvmma/buddy32.h>
#include <string.h>
#include <stdlib.h>

#define EXCVEC_CAPA          16
#define TYPEVEC_CAPA         16

/*----------------------------------------------------------------*/

wt_vmptr wtr_vm_vma_malloc(struct wtr_vm* const vm, size_t size)
{
	const wt_vmptr vma = buddy32v_alloc(vm->heap_buddy, size);

	if(0!=vma && 0!=wtr_vma_acquire(vm, vma, size))
	{
		buddy32v_free(vm->heap_buddy, vma);
		return 0;
	}

	return vma;
}

void wtr_vm_vma_free(struct wtr_vm* const vm, wt_vmptr memory)
{
	struct buddy32v* const buddy = vm->heap_buddy;

	const wt_u32 size = buddy32v_msize(buddy, memory);

	if(0==buddy32v_free(buddy, memory))
	{
		wtr_vma_release(vm, memory, size);
	}
}

int wtr_vma_acquire(struct wtr_vm* const vm, wt_vmptr vma, wt_u32 size)
{
	const wt_u32  szsh   = vm->vma_segsh;
	wt_u32* const segacq = vm->vma_segacq;

	wt_vmptr start_seg = vma >> szsh;
	wt_vmptr end_seg   = (vma + size - 1) >> szsh;

	for(;start_seg<=end_seg;++start_seg)
	{
		if(0==segacq[start_seg])
		{
			/* TODO: we could check whether this failed or not */
			vmmu32_acquire_segment(&vm->vmmu,
				&vm->vmmu.segments[start_seg]);
		}

		++segacq[start_seg];
	}

	return 0;
}

int wtr_vma_release(struct wtr_vm* const vm, wt_vmptr vma, wt_u32 size)
{
	const wt_u32  szsh   = vm->vma_segsh;
	wt_u32* const segacq = vm->vma_segacq;

	wt_vmptr start_seg = vma >> szsh;
	wt_vmptr end_seg   = (vma + size - 1) >> szsh;

	for(;start_seg<=end_seg;++start_seg)
	{
		--segacq[start_seg];

		if(0==segacq[start_seg])
		{
			/* TODO: we could check whether this failed or not */
			vmmu32_release_segment(&vm->vmmu,
				&vm->vmmu.segments[start_seg]);
		}
	}

	return 0;
}

void* wtr_vm_sys_realloc(struct wtr_vm* const vm, void* memory,
	const size_t old_size, const size_t new_size)
{
	void* const  alloc_ctx  = vm->alloc_ctx;
	WT_PFN_ALLOC alloc_func = vm->alloc_func;
	void*        new_mem;

	if(NULL==memory)
	{
		return alloc_func(alloc_ctx, NULL, new_size);
	}

	if(old_size==new_size)
	{
		/* this should not happen, but since they are equal do nothing */
		return memory;
	}

	if(NULL!=(new_mem=alloc_func(alloc_ctx, memory, new_size)))
	{
		return new_mem;
	}

	/* alloc_func() could not reallocate the block, but this could be */
	/* because it does not implement realloc, so try a different approach */

	if(NULL==(new_mem=alloc_func(alloc_ctx, NULL, new_size)))
	{
		/* nope, we sure ran out of memory */
		return NULL;
	}

	if(old_size>new_size)
	{
		memcpy(new_mem, memory, new_size);
	}
	else
	{
		memcpy(new_mem, memory, old_size);
	}

	alloc_func(alloc_ctx, memory, 0);

	return new_mem;
}

void* wtr_vm_sys_malloc(struct wtr_vm* const vm, size_t size)
{
	return vm->alloc_func(vm->alloc_ctx, NULL, size);
}

void wtr_vm_sys_free(struct wtr_vm* const vm, void* memory)
{
	vm->alloc_func(vm->alloc_ctx, memory, 0);
}

struct wtr_exception* wtr_vm_makeexception(struct wtr_vm* const vm,
	struct wtr_program* prog, const char* str)
{
	const size_t len = strlen(str) + 1;

	struct wtr_exception* const exc =
		(struct wtr_exception*)wtr_vm_sys_malloc(vm,
			sizeof(struct wtr_exception)+len*sizeof(*str));

	if(NULL!=exc)
	{
		exc->refcount  = 1;
		exc->exception = 0;
		exc->owner     = prog;
		memcpy(exc->name, str, len*sizeof(*str));
	}

	return exc;
}

struct wtr_exception* wtr_vm_findexception(struct wtr_vm* const vm,
	const char* const excname)
{
	const size_t size = wtr_vector_size(&vm->exception_vec);

	if(0<size)
	{
		struct wtr_exception** ptr = (struct wtr_exception**)
			wtr_vector_at(&vm->exception_vec, 0);

		struct wtr_exception** const end = ptr + size;

		while(ptr<end)
		{
			struct wtr_exception* const exc = *ptr++;

			if(NULL!=exc && 0==strcmp(exc->name, excname))
			{
				return exc;
			}
		}
	}

	return NULL;
}

wt_bool wtr_vm_removeexception(struct wtr_vm* const vm,
	struct wtr_exception* const exc)
{
	size_t size = wtr_vector_size(&vm->exception_vec);

	if(0<exc->refcount || exc->exception>=size)
	{
		return WT_FALSE;
	}

	wtr_vector_set(&vm->exception_vec, exc->exception, NULL);

	memset(exc, 0, sizeof(struct wtr_exception));
	wtr_vm_sys_free(vm, exc);

	/* pop trailing NULLs, but the vector must have 1 remaining element */

	while(1<(size=wtr_vector_size(&vm->exception_vec)) &&
		NULL==wtr_vector_at(&vm->exception_vec, size-1))
	{
		wtr_vector_pop_back(&vm->exception_vec);
	}

	return WT_TRUE;
}

wt_error wtr_vm_addexception(struct wtr_vm* const vm,
	struct wtr_program* const prog, const char* const excname,
	struct wtr_exception** const p_exc)
{
	struct wtr_exception* exception;

	if(NULL!=(exception=wtr_vm_findexception(vm, excname)))
	{
		*p_exc = exception;
		++exception->refcount;
		return WTE_OK;
	}
	else
	{
		if(NULL==(exception=wtr_vm_makeexception(vm, prog, excname)))
		{
			return WTE_NOMEM;
		}
		else
		{
			wt_exception exc_id = 0;
			char*        name   = exception->name;
			wt_bool      push   = WT_TRUE;
			const size_t size   = wtr_vector_size(&vm->exception_vec);

			for(;exc_id<size;++exc_id)
			{
				if(NULL==wtr_vector_at(&vm->exception_vec, exc_id))
				{
					push = WT_FALSE;
					break;
				}
			}

			if(
				(push && 0!=wtr_vector_push_back(&vm->exception_vec, vm,
					exception))
				||
				(!push && 0!=wtr_vector_set(&vm->exception_vec, exc_id,
					exception))
			)
			{
				wtr_vm_sys_free(vm, exception);
				return WTE_NOMEM;
			}

			*p_exc = exception;
			exception->exception = exc_id;
		}
	}

	return WTE_OK;
}

struct wtr_function* wtr_vm_findfunction(struct wtr_vm* const vm,
	struct wtr_funcsig* const sig)
{
	struct wtr_list* const list = &vm->list_functions;
	struct wtr_node*       it;

	wtr_list_for_each(list, it)
	{
		struct wtr_function* const func = wtr_contof(it, struct wtr_function,
			link_functions);

		if(0==wtr_funcsig_cmp(sig, func->funcsig))
		{
			return func;
		}
	}

	return NULL;
}

struct wtr_type* wtr_vm_findtype(struct wtr_vm* const vm,
	const char* const name)
{
	const size_t size = wtr_vector_size(&vm->type_vec);

	if(0<size)
	{
		struct wtr_type** ptr = (struct wtr_type**)
			wtr_vector_at(&vm->type_vec, 0);

		struct wtr_type** const end = ptr + size;

		while(ptr<end)
		{
			struct wtr_type* const typ = *ptr++;

			if(NULL!=typ && 0==strcmp(typ->name, name))
			{
				return typ;
			}
		}
	}

	return NULL;
}

wt_bool wtr_vm_addtype(struct wtr_vm* const vm,
	struct wtr_type* const type)
{
	wt_u32 idx = 0;
	wt_u32 num = wtr_vector_size(&vm->type_vec);

	/* search an unused index */

	{
		struct wtr_type** ptr = (struct wtr_type**)
			wtr_vector_at(&vm->type_vec, 0);

		for(idx=0;idx!=num;++idx)
		{
			if(NULL==(*ptr))
			{
				break;
			}

			++ptr;
		}
	}

	type->id = idx;

	if(idx==num)
	{
		if(0!=wtr_vector_push_back(&vm->type_vec, vm, type))
		{
			return WT_FALSE;
		}
	}
	else
	{
		if(0!=wtr_vector_set(&vm->type_vec, idx, type))
		{
			return WT_FALSE;
		}
	}

	return WT_TRUE;
}

void wtr_vm_removetype(struct wtr_vm* const vm,
	struct wtr_type* const type)
{
	wtr_vector_set(&vm->type_vec, type->id, NULL);
	wtr_type_release(type);
}

struct wtr_operator* wtr_vm_findoperator(struct wtr_vm* const vm,
	const char* const name)
{
	struct wtr_list* const list = &vm->list_operators;
	struct wtr_node*       it;

	wtr_list_for_each(list, it)
	{
		struct wtr_operator* const op = wtr_contof(it, struct wtr_operator,
			link_ops);

		if(0==strcmp(name, op->name))
		{
			return op;
		}
	}

	return NULL;
}

wt_bool wtr_vm_addoperator(struct wtr_vm* const vm,
	struct wtr_operator* const op)
{
	if(NULL!=wtr_vm_findoperator(vm, op->name))
	{
		return WT_FALSE;
	}

	wtr_list_insert_tail(&vm->list_operators, &op->link_ops);

	return WT_TRUE;
}

void wtr_vm_removeoperator(struct wtr_vm* const vm,
	struct wtr_operator* const op)
{
	wtr_list_remove(&op->link_ops);
	wtr_list_init(&op->link_ops);
	wtr_operator_release(op);
}

size_t wtr_sizeof(wt_u32 type)
{
	if(type & (WTTF_SEQ | WTTF_REF | WTTF_LOCAL))
	{
		return sizeof(wt_vmptr);
	}

	if(WTT_XSTRING==type)
	{
		return sizeof(wt_vmptr);
	}

	switch(type & WTTM_BASICTYPE)
	{
	case WTT_BOOL:    return sizeof(wt_bool);
	case WTT_STRING:  return sizeof(wt_vmptr);  /* this is an object */
	case WTT_I32:     return sizeof(wt_i32);
	case WTT_U32:     return sizeof(wt_u32);
	case WTT_I64:     return sizeof(wt_i64);
	case WTT_U64:     return sizeof(wt_u64);
#if WTRUNTIME_ENABLE_FLOAT32
	case WTT_F32:     return sizeof(wt_f32);
#endif /* WTRUNTIME_ENABLE_FLOAT32 */
#if WTRUNTIME_ENABLE_FLOAT64
	case WTT_F64:     return sizeof(wt_f64);
#endif /* WTRUNTIME_ENABLE_FLOAT64 */
	case WTT_CHAR:    return sizeof(wt_char);
	case WTT_HANDLE:  return sizeof(wt_handle);
	case WTT_FILE:    return sizeof(wt_handle); /* this is a pointer, but it will fit in this */
	case WTT_DIR:     return sizeof(wt_handle); /* this is a pointer, but it will fit in this */
	case WTT_PROGRAM: return sizeof(wt_handle); /* this is a pointer, but it will fit in this */

	case WTT_I8:
	case WTT_U8:
	case WTT_I16:
	case WTT_U16:
		return sizeof(wt_u32);
	}

	return 0;
}

/*---------------------------------------------------------------- */

static void release_vmmu_segments(struct wtr_vm* const vm)
{
	wt_u32 i;

	for(i=0;i<vm->vmmu.nsegments;++i)
	{
		vmmu32_release_segment(&vm->vmmu, &vm->vmmu.segments[i]);
	}
}

wt_error wtrCreateTombVM(struct wtr_vm** const p_vm, wt_u32 flags,
	struct vmmu32_segment* const vma_segs, size_t vma_nsegs, wt_u32 vma_szsh,
	wt_u32 heap_start, wt_u32 heap_size, wt_u32 heap_gran, wt_u8 ratio_css,
	wt_u8 ratio_dss, void* alloc_ctx, WT_PFN_ALLOC alloc_func)
{
	struct wtr_vm* vm        = NULL;
	wt_u32         acc_flags = 0; /* accepted flags (WTRVMF_*) */

	if(NULL==p_vm || 0==ratio_css || 0==ratio_dss || NULL==alloc_func)
	{
		return WTE_BADARG;
	}

	if(flags & WTRVMF_REQINTEGRITY)
	{
		acc_flags |= WTRVMF_REQINTEGRITY;
	}

	if(NULL==(vm=(struct wtr_vm*)alloc_func(alloc_ctx, NULL,
		sizeof(struct wtr_vm))))
	{
		return WTE_NOMEM;
	}

	memset(vm, 0, sizeof(struct wtr_vm));

	vm->alloc_ctx      = alloc_ctx;
	vm->alloc_func     = alloc_func;
	vm->ratio_css      = ratio_css;
	vm->ratio_dss      = ratio_dss;
	vm->accepted_flags = acc_flags;
	vm->vma_segsh      = vma_szsh;
	vm->xs_x           = (wt_u32)(wt_uintptr)vm;
	vm->xs_y           = (wt_u32)(wt_uintptr)wtrCreateTombVM;
	vm->xs_z           = 0x5a5a5a5a ^ vm->xs_x ^ vm->xs_y;
	vm->xs_w           = 0xa5a5a5a5 ^ vm->xs_x ^ vm->xs_y;

	wtr_list_init(&vm->list_programs);
	wtr_list_init(&vm->list_functions);
	wtr_list_init(&vm->list_operators);

	{
		const int ec = -vmmu32_init(&vm->vmmu, vma_segs, vma_nsegs, vma_szsh);

		switch(ec)
		{
		case 0:
			/* success */
			break;

		case VMMU_ERR_EINVAL:
			return WTE_VINVAL;

		case VMMU_ERR_ENSEGS:
			return WTE_VNSEGS;

		case VMMU_ERR_ESEGSIZE:
			return WTE_VSEGSZ;

		default:
			return WTE_VINIT;
		}
	}

	// TODO: apply limits on 'heap_gran'?
	{
		const size_t state_size = buddy32v_state_size(heap_size, heap_gran);

		const size_t total_size =
			state_size + sizeof(vm->vma_segacq[0])*vma_nsegs;

		if(0==state_size)
		{
			alloc_func(alloc_ctx, vm, 0);
			return WTE_BADHEAP;
		}

		if(NULL==(vm->heap_sysmem=alloc_func(alloc_ctx, NULL, total_size)))
		{
			alloc_func(alloc_ctx, vm, 0);
			return WTE_NOMEM;
		}

		memset(vm->heap_sysmem, 0, total_size);
		vm->vma_segacq = (wt_u32*)(((wt_u8*)vm->heap_sysmem) + state_size);

		if(NULL==(vm->heap_buddy=buddy32v_init(vm->heap_sysmem, state_size,
			heap_start, heap_size, heap_gran)))
		{
			alloc_func(alloc_ctx, vm->heap_sysmem, 0);
			alloc_func(alloc_ctx, vm, 0);
			return WTE_BADHEAP;
		}
	}

	if(0!=wtr_vector_create(&vm->type_vec, vm, TYPEVEC_CAPA))
	{
		goto error_1;
	}

	if(0!=wtr_vector_create(&vm->exception_vec, vm, EXCVEC_CAPA))
	{
		goto error_2;
	}

	if(0==(vm->pc_tcl_ja=wtr_vm_vma_malloc(vm, sizeof(wtr_routine_tcl_ja))))
	{
		goto error_3;
	}
	else
	{
		if(0!=wtr_vma_write(vm, vm->pc_tcl_ja, wtr_routine_tcl_ja,
			sizeof(wtr_routine_tcl_ja)))
		{
			goto error_3;
		}

		vm->pc_tcl_ja /= sizeof(wt_u16);
	}

	if(0==(vm->pc_tcl_leave=wtr_vm_vma_malloc(vm,
		sizeof(wtr_routine_tcl_leave))))
	{
		goto error_3;
	}
	else
	{
		if(0!=wtr_vma_write(vm, vm->pc_tcl_leave, wtr_routine_tcl_leave,
			sizeof(wtr_routine_tcl_leave)))
		{
			goto error_3;
		}

		vm->pc_tcl_leave /= sizeof(wt_u16);
	}

	if(0==(vm->pc_strd_leave=wtr_vm_vma_malloc(vm,
		sizeof(wtr_routine_strd_leave))))
	{
		goto error_3;
	}
	else
	{
		if(0!=wtr_vma_write(vm, vm->pc_strd_leave, wtr_routine_strd_leave,
			sizeof(wtr_routine_strd_leave)))
		{
			goto error_3;
		}

		vm->pc_strd_leave /= sizeof(wt_u16);
	}

	if(0==(vm->pc_cfp_leave=wtr_vm_vma_malloc(vm,
		sizeof(wtr_routine_cfp_leave))))
	{
		goto error_3;
	}
	else
	{
		if(0!=wtr_vma_write(vm, vm->pc_cfp_leave, wtr_routine_cfp_leave,
			sizeof(wtr_routine_cfp_leave)))
		{
			goto error_3;
		}

		vm->pc_cfp_leave /= sizeof(wt_u16);
	}

	vm->external_tag = ~((wt_u32)0);

	{
		struct wtr_exception* ex;

#define WTR_VM_ADDEXCEPTION(exc) \
		if(WTE_OK!=wtr_vm_addexception(vm, NULL, #exc, &ex)) \
		{ \
			goto error_4; \
		}

		WTR_STDEXCEPTIONS(WTR_VM_ADDEXCEPTION);
	}

	*p_vm = vm;
	return WTE_OK;

error_4:;
	{
		size_t size = wtr_vector_size(&vm->exception_vec);

		while(0<size)
		{
			struct wtr_exception** const p_exc =
				(struct wtr_exception**)wtr_vector_at(&vm->exception_vec,
					--size);

			wtr_vm_sys_free(vm, *p_exc);
		}
	}
error_3:;
	wtr_vector_delete(&vm->exception_vec, vm);
error_2:;
	wtr_vector_delete(&vm->type_vec, vm);
error_1:
	release_vmmu_segments(vm);
	alloc_func(alloc_ctx, vm->heap_sysmem, 0);
	alloc_func(alloc_ctx, vm, 0);
	return WTE_NOMEM;
}

wt_error wtrVMDestroy(struct wtr_vm* const vm)
{
	size_t size;

	WT_PFN_ALLOC alloc_func = vm->alloc_func;
	void*        alloc_ctx  = vm->alloc_ctx;

	if(NULL==vm)
	{
		return WTE_BADARG;
	}

	if(0<vm->num_context)
	{
		return WTE_REFCTX;
	}

	while(0<(size=wtr_vector_size(&vm->exception_vec)))
	{
		struct wtr_exception* exc;

		struct wtr_exception** const p_exc =
			(struct wtr_exception**)wtr_vector_at(&vm->exception_vec,
				size-1);

		if(NULL==p_exc || NULL==(exc=(*p_exc)))
		{
			wtr_vector_pop_back(&vm->exception_vec);
			continue;
		}

		if(NULL!=exc->owner || 1!=exc->refcount)
		{
			return WTE_REFER;
		}

		--exc->refcount;

		if(!wtr_vm_removeexception(vm, exc))
		{
			return WTE_REFER;
		}
	}

	wtr_vector_delete(&vm->exception_vec, vm);
	wtr_vector_delete(&vm->type_vec, vm);

	release_vmmu_segments(vm);

	alloc_func = vm->alloc_func;
	alloc_ctx  = vm->alloc_ctx;

	alloc_func(alloc_ctx, vm->heap_sysmem, 0);
	alloc_func(alloc_ctx, vm, 0);

	return WTE_OK;
}

wt_error wtrVMSetLoaderOps(struct wtr_vm* const vm, void* const ops_ctx,
	const struct wtr_loaderops* const ops)
{
	vm->loaderops_ctx = ops_ctx;
	vm->loaderops     = ops;

	return WTE_OK;
}

wt_error wtrVMSetHashAlgs(struct wtr_vm* const vm,
	const struct wtr_hashalg* const algs)
{
	vm->hashalgs = algs;

	return WTE_OK;
}

wt_error wtrVMGetException(struct wtr_vm* const vm, const char* const name,
	wt_exception* const p_exc)
{
	if(NULL==name || NULL==p_exc)
	{
		return WTE_BADARG;
	}
	else
	{
		struct wtr_exception* const exception =
			wtr_vm_findexception(vm, name);

		if(NULL==exception)
		{
			return WTE_NOENT;
		}

		*p_exc = exception->exception;
	}

	return WTE_OK;
}

wt_error wtrVMExceptionToString(struct wtr_vm* const vm, wt_exception exc,
	char* namebuf, size_t bufsize)
{
	if(NULL==namebuf)
	{
		return WTE_BADARG;
	}
	else
	{
		struct wtr_exception** const p_exc =
			(struct wtr_exception**)wtr_vector_at(&vm->exception_vec, exc);

		size_t len;

		if(NULL==p_exc)
		{
			return WTE_NOENT;
		}

		len = strlen((*p_exc)->name);

		if(len>=bufsize)
		{
			return WTE_NOFIT;
		}

		memcpy(namebuf, (*p_exc)->name, len+1);
	}

	return WTE_OK;
}

wt_error wtrVMAddException(struct wtr_vm* const vm, const char* const name,
	wt_exception* const out_exc)
{
	struct wtr_exception* ex;
	const wt_error err = wtr_vm_addexception(vm, NULL, name, &ex);

	if(WTE_OK==err && NULL!=out_exc)
	{
		*out_exc = ex->exception;
	}

	return err;
}

wt_error wtrVMAddFunction(struct wtr_vm* const vm, const char* name,
	wt_u32 argc, const wt_u32* argv, wt_u32 retc, const wt_u32* retv,
	WT_PFN_EXTFUNC func, void* token)
{
	if(NULL==name || NULL==func || !(argc==0 || argv) || !(retc==0 || retv))
	{
		return WTE_BADARG;
	}
	else
	{
		struct wtr_funcsig* const sig = wtr_funcsig_new(vm, name, argc, argv,
			retc, retv, WT_FALSE, NULL, wtr_byteorder());

		if(NULL==sig)
		{
			return WTE_NOMEM;
		}

		if(NULL!=wtr_vm_findfunction(vm, sig))
		{
			wtr_funcsig_delete(vm, sig);
			return WTE_DUPENT;
		}
		else
		{
			struct wtr_cfunction* const cf = wtr_cfunction_new(vm, sig, func,
				token);

			if(NULL==cf)
			{
				wtr_funcsig_delete(vm, sig);
				return WTE_NOMEM;
			}
			else
			{
				struct wtr_function* const fh = wtr_function_new(vm, sig, cf,
					NULL, 0);

				if(NULL==fh)
				{
					wtr_cfunction_delete(vm, cf);
					wtr_funcsig_delete(vm, sig);
					return WTE_NOMEM;
				}
			}
		}
	}

	return WTE_OK;
}

wt_error wtrVMGetFunction(struct wtr_vm* const vm, wt_function* p_func,
	const char* name, wt_u32 argc, const wt_u32* argv, wt_u32 retc,
	const wt_u32* retv, wt_bool pure)
{
	if(NULL==p_func || NULL==name)
	{
		return WTE_BADARG;
	}
	else
	{
		struct wtr_funcsig* const sig = wtr_funcsig_new(vm, name, argc, argv,
			retc, retv, pure, NULL, wtr_byteorder());

		if(NULL==sig)
		{
			return WTE_NOMEM;
		}
		else
		{
			struct wtr_function* const func =
				wtr_vm_findfunction(vm, sig);

			if(NULL==func)
			{
				return WTE_NOENT;
			}
			else
			{
				*p_func = wtr_function_genhandle(vm, func);
				wtr_funcsig_delete(vm, sig);

				if(NULL==(*p_func))
				{
					return WTE_NOMEM;
				}
			}
		}
	}

	return WTE_OK;
}

wt_error wtrVMReleaseFunction(struct wtr_vm* const vm, wt_function func)
{
	if(NULL==func)
	{
		return WTE_BADARG;
	}

	wtr_function_remhandle(vm, func);
	return WTE_OK;
}

wt_error wtrVMLoadProgram(struct wtr_vm* const vm, wt_program* program,
	struct wtr_context* const ctx, const void* data, size_t size,
	wt_u32 flags)
{
	if(NULL==program || NULL==data || 0==size || NULL==ctx)
	{
		return WTE_BADARG;
	}

	if(WTRR_INIT!=ctx->run)
	{
		return WTE_BADCTX;
	}

	return wtr_loader(vm, program, ctx, data, size, flags);
}

wt_error wtrVMReleaseProgram(struct wtr_vm* const vm, wt_program program,
	struct wtr_context* const ctx)
{
	size_t i;

	if(NULL==vm || NULL==program || NULL==ctx || vm!=program->vm)
	{
		return WTE_BADARG;
	}

	if(WTRR_INIT!=ctx->run)
	{
		return WTE_BADCTX;
	}

	if(1<program->refcount)
	{
		return WTE_REFER;
	}
	else
	if(0==program->refcount)
	{
		return WTE_DESTRP;
	}

	if(0<wtr_vector_size(&program->own_exceptions))
	{
		struct wtr_exception** ptr = (struct wtr_exception**)
			wtr_vector_at(&program->own_exceptions, 0);

		struct wtr_exception** const end = ptr +
			wtr_vector_size(&program->own_exceptions);

		while(ptr<end)
		{
			struct wtr_exception* const exc = *ptr++;

			/* the refcount can be at most 1 (reference counter will be */
			/* increased by 1 when the program imports it) */

			if(1<exc->refcount)
			{
				return WTE_REFER;
			}
		}
	}

	if(0<wtr_vector_size(&program->own_types))
	{
		struct wtr_type** ptr = (struct wtr_type**)
			wtr_vector_at(&program->own_types, 0);

		struct wtr_type** const end = ptr +
			wtr_vector_size(&program->own_types);

		while(ptr<end)
		{
			struct wtr_type* const typ = *ptr++;

			/* the refcount can be at most 1 (reference counter will be */
			/* increased by 1 when the program imports it) */

			if(1<typ->refcount)
			{
				return WTE_REFER;
			}
		}
	}

	if(0<wtr_vector_size(&program->own_operators))
	{
		struct wtr_operator** ptr = (struct wtr_operator**)
			wtr_vector_at(&program->own_operators, 0);

		struct wtr_operator** const end = ptr +
			wtr_vector_size(&program->own_operators);

		while(ptr<end)
		{
			struct wtr_operator* const op = *ptr++;

			/* the refcount shall be zero now */

			if(0<op->refcount)
			{
				return WTE_REFER;
			}
		}
	}

	--program->refcount;

	if(program->pc_dtor)
	{
		wtr_ctx_invoke_dtor(ctx, program);
	}
	else
	{
		wtr_program_destroy(vm, program);
	}

	return WTE_OK;
}

wt_error wtrVMSetUserPointer(struct wtr_vm* const vm, void* ptr)
{
	vm->user_ptr = ptr;
	return WTE_OK;
}

wt_error wtrVMGetUserPointer(struct wtr_vm* const vm, void** pptr)
{
	*pptr = vm->user_ptr;
	return WTE_OK;
}

int wtr_vma_memcpy(struct wtr_vm* const vm, wt_vmptr dst, wt_vmptr src,
	wt_u32 size)
{
	wt_u8 buf[32];

	while(sizeof(buf)<=size)
	{
		if(0!=vmmu32_read(&vm->vmmu, src, buf, sizeof(buf)))
		{
			return 1;
		}

		if(0!=vmmu32_write(&vm->vmmu, dst, buf, sizeof(buf)))
		{
			return 2;
		}

		dst  += sizeof(buf);
		src  += sizeof(buf);
		size -= sizeof(buf);
	}

	if(0<size)
	{
		if(0!=vmmu32_read(&vm->vmmu, src, buf, size))
		{
			return 3;
		}

		if(0!=vmmu32_write(&vm->vmmu, dst, buf, size))
		{
			return 4;
		}
	}

	return 0;
}

int wtr_vma_memmove(struct wtr_vm* const vm, wt_vmptr dst, wt_vmptr src,
	wt_u32 size)
{
	wt_u8 buf[32];

	if(dst<src)
	{
		return wtr_vma_memcpy(vm, dst, src, size);
	}

	src += size;
	dst += size;

	while(sizeof(buf)<=size)
	{
		dst -= sizeof(buf);
		src -= sizeof(buf);

		if(0!=vmmu32_read(&vm->vmmu, src, buf, sizeof(buf)))
		{
			return 1;
		}

		if(0!=vmmu32_write(&vm->vmmu, dst, buf, sizeof(buf)))
		{
			return 2;
		}

		size -= sizeof(buf);
	}

	if(0<size)
	{
		dst -= size;
		src -= size;

		if(0!=vmmu32_read(&vm->vmmu, src, buf, size))
		{
			return 3;
		}

		if(0!=vmmu32_write(&vm->vmmu, dst, buf, size))
		{
			return 4;
		}
	}

	return 0;
}

int wtr_vma_memset(struct wtr_vm* const vm, wt_vmptr vma, int c,
	wt_u32 size)
{
	wt_u8 buf[32];

	memset(buf, c, sizeof(buf));

	while(sizeof(buf)<=size)
	{
		if(0!=vmmu32_write(&vm->vmmu, vma, buf, sizeof(buf)))
		{
			return 1;
		}

		vma  += sizeof(buf);
		size -= sizeof(buf);
	}

	if(0<size)
	{
		if(0!=vmmu32_write(&vm->vmmu, vma, buf, size))
		{
			return 4;
		}
	}

	return 0;
}

int wtr_vma_memcmp(struct wtr_vm* const vm, wt_vmptr s1, wt_vmptr s2,
	wt_u32 size)
{
	wt_u8 buf1[16];
	wt_u8 buf2[sizeof(buf1)];

	while(sizeof(buf1)<=size)
	{
		int res;

		if(0!=vmmu32_read(&vm->vmmu, s1, buf1, sizeof(buf1)))
		{
			return 1;
		}

		if(0!=vmmu32_read(&vm->vmmu, s2, buf2, sizeof(buf2)))
		{
			return 2;
		}

		if(0!=(res=memcmp(buf1, buf2, sizeof(buf1))))
		{
			return res;
		}

		s1   += sizeof(buf1);
		s2   += sizeof(buf1);
		size -= sizeof(buf1);
	}

	if(0<size)
	{
		if(0!=vmmu32_read(&vm->vmmu, s1, buf1, size))
		{
			return 3;
		}

		if(0!=vmmu32_read(&vm->vmmu, s2, buf2, size))
		{
			return 4;
		}

		return memcmp(buf1, buf2, size);
	}

	return 0;
}

int wtr_vma_strlen(struct wtr_vm* const vm, wt_vmptr str, wt_u32* const len)
{
	const wt_vmptr base = str;

	wt_u8 ch;

	do
	{
		if(0!=wtr_vma_read(vm, str++, &ch, sizeof(ch)))
		{
			return 1;
		}
	}
	while(0!=ch);

	*len = (str-1) - base;

	return 0;
}

wt_error wtrVMReadMemory(struct wtr_vm* const vm, wt_vmptr vma,
	void* ptr, wt_u32 size)
{
	if(0!=vmmu32_read(&vm->vmmu, vma, ptr, size))
	{
		return WTE_FAULT;
	}

	return WTE_OK;
}

wt_error wtrVMWriteMemory(struct wtr_vm* const vm, wt_vmptr vma,
	const void* ptr, wt_u32 size)
{
	if(0!=vmmu32_write(&vm->vmmu, vma, ptr, size))
	{
		return WTE_FAULT;
	}

	return WTE_OK;
}
